"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OfflineMirrorRegistry = void 0;
const tslib_1 = require("tslib");
const express_1 = tslib_1.__importDefault(require("express"));
require("express-async-errors");
const morgan_1 = tslib_1.__importDefault(require("morgan"));
const glob_1 = tslib_1.__importDefault(require("glob"));
const node_cache_1 = tslib_1.__importDefault(require("node-cache"));
const tar_1 = tslib_1.__importDefault(require("tar"));
const fs_1 = require("fs");
const minimatch_1 = tslib_1.__importDefault(require("minimatch"));
const ssri_1 = tslib_1.__importDefault(require("ssri"));
const path_1 = require("path");
// OfflineMirrorRegistry
class OfflineMirrorRegistry {
    constructor(props) {
        this.cache = props.cache || new node_cache_1.default();
        this.directory = props.directory || ".";
        this.useHostname = props.useHostname || false;
    }
    // Find packages for this name in the specified directory.
    getPackages(dir, name) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            // Check if it's in the cache
            const maybeCached = this.cache.get(`pkg:${name}`);
            if (maybeCached) {
                return maybeCached;
            }
            // Get the file prefix
            const filePrefix = this.transformName(name);
            // Create a list of all results we find...
            const results = {};
            // Find all files that match this package name
            const matched = glob_1.default.sync(`${filePrefix}-+([0-9]).+([0-9]).+([0-9]).tgz`, {
                cwd: dir,
                nodir: true,
                absolute: true,
            });
            if (matched.length === 0) {
                // nothing matched!
                console.warn(`found no results for ${name} (with prefix ${filePrefix})`);
                return null;
            }
            // Go through each match, untar, figure out whether this is what we want.
            for (const match of matched) {
                console.log(match);
                let bufs = Buffer.allocUnsafe(0); // buffer for package.json
                let unpackedSize = 0;
                let fileCount = 0; // includes directories!
                yield tar_1.default.t({
                    onentry: (entry) => { var entry_1, entry_1_1; return tslib_1.__awaiter(this, void 0, void 0, function* () {
                        var e_1, _a;
                        fileCount++;
                        let packageJson = false;
                        if (entry.type === "File" &&
                            (0, minimatch_1.default)(entry.path, "*/package.json")) {
                            packageJson = true;
                        }
                        try {
                            // We MUST consume the data no matter what; it's whether it's package.json
                            // is whether we store it.
                            for (entry_1 = tslib_1.__asyncValues(entry); entry_1_1 = yield entry_1.next(), !entry_1_1.done;) {
                                const data = entry_1_1.value;
                                unpackedSize += data.length;
                                if (packageJson) {
                                    bufs = Buffer.concat([bufs, data]);
                                }
                            }
                        }
                        catch (e_1_1) { e_1 = { error: e_1_1 }; }
                        finally {
                            try {
                                if (entry_1_1 && !entry_1_1.done && (_a = entry_1.return)) yield _a.call(entry_1);
                            }
                            finally { if (e_1) throw e_1.error; }
                        }
                    }); },
                    file: match,
                });
                const integrity = yield ssri_1.default.fromStream((0, fs_1.createReadStream)(match), {
                    algorithms: ["sha512"],
                });
                if (bufs.length === 0) {
                    // We didn't find a package.json? Weird. Skip.
                    console.warn(`${name}: tried matching ${match} but couldn't find package.json`);
                    continue;
                }
                const packageJsonStr = bufs.toString("utf8");
                const packageJson = JSON.parse(packageJsonStr);
                if (packageJson.name !== name) {
                    // picked up the wrong package
                    console.warn(`${name}: tried matching ${match} but found different name in package.json`);
                    continue;
                }
                // Excellent, build the record
                results[packageJson.version] = {
                    name: packageJson.name,
                    version: packageJson.version,
                    dependencies: packageJson.dependencies,
                    devDependencies: packageJson.devDependencies,
                    dist: {
                        fileCount,
                        unpackedSize,
                        integrity: integrity.toString(),
                        tarball: match,
                    },
                };
            }
            // We found nothing.
            if (Object.keys(results).length === 0) {
                console.warn(`matched files for ${name} (with prefix ${filePrefix}) but no correct package.json?`);
                return null;
            }
            // We found something!
            const packagesRecord = {
                name,
                versions: results,
            };
            // Cache the record...
            this.cache.set(`pkg:${name}`, packagesRecord);
            // Then cache all the tarball names to their absolute paths for good measure
            for (const result of Object.values(results)) {
                this.cache.set(`tar:${name}:${(0, path_1.basename)(result.dist.tarball)}`, result.dist.tarball);
            }
            // Return the created record
            return packagesRecord;
        });
    }
    getServer() {
        // Set up the registry mimic
        const app = (0, express_1.default)();
        // Set up the logger
        app.use((0, morgan_1.default)("tiny"));
        // Set up package functions
        app.get("/:package([^/]+((/[^/]+)?))", (req, res) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            // Get the package name
            const pkg = req.params.package;
            // Get the host
            const proto = req.protocol;
            const hostname = this.useHostname ? req.hostname : req.headers.host;
            // Try to get the package
            const packagesRecord = yield this.getPackages(this.directory, pkg);
            if (!packagesRecord) {
                // 404
                res.status(404).json([]);
                return;
            }
            // Copy the record
            const packagesRecordCopy = structuredClone(packagesRecord);
            // and modify all tarballs to use the proto/host/port
            for (const version of Object.values(packagesRecordCopy.versions)) {
                version.dist.tarball = `${proto}://${hostname}/${pkg}/-/${(0, path_1.basename)(version.dist.tarball)}`;
            }
            // Return results
            res.json(packagesRecordCopy);
        }));
        // Set up tarball retrieval
        app.get("/:package([^/]+((/[^/]+)?))/-/:tarball", (req, res) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            // Get the package name and the wanted tarball
            const pkg = req.params.package;
            const tarball = req.params.tarball;
            // make sure we're cached...
            yield this.getPackages(this.directory, pkg);
            // and see if there's a tarball
            const cachedPath = this.cache.get(`tar:${pkg}:${tarball}`);
            if (!cachedPath) {
                // This does not exist, die.
                res.status(404).send();
                return;
            }
            // Send the file!
            res.sendFile(cachedPath);
        }));
        // Anything else, 404
        app.use((req, res) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            res.status(404).json({});
        }));
        return app;
    }
    // Follow yarn offline cache's method - which is just replacing / with -
    transformName(name) {
        return name.replace(/\//g, "-");
    }
}
exports.OfflineMirrorRegistry = OfflineMirrorRegistry;
//# sourceMappingURL=index.js.map